// Copyright 2025 The Go Authors. All rights reserved.
// Use of this source code is governed by a BSD-style
// license that can be found in the LICENSE file.

//go:build goexperiment.ms_tls_config_schannel

package tls

import (
	"bytes"
	"os/exec"
	"sync"
	"testing"
)

var schannelSuites = sync.OnceValue(func() []uint16 {
	cipherSuites, err := cipherSuitesSchannel()
	if err != nil {
		panic(err) // This should never happen, as Schannel is always available on Windows.
	}
	return cipherSuites
})

func TestCipherSuitesSchannel(t *testing.T) {
	cipherSuites, err := cipherSuitesSchannel()
	if err != nil {
		t.Fatal(err)
	}
	if len(cipherSuites) == 0 {
		t.Fatal("cipherSuitesSchannel returned no cipher suites")
	}

	for _, suite := range cipherSuites {
		if suite == 0 {
			t.Fatal("cipherSuitesSchannel returned a 0 cipher suite")
		}
	}

	// Check that all the cipher suites are present in the Get-TlsCipherSuite output.
	output, err := exec.Command("powershell", "-Command", "Get-TlsCipherSuite").CombinedOutput()
	if err != nil {
		t.Fatalf("failed to get TLS cipher suites: %v\n%s", err, output)
	}
	for _, id := range cipherSuites {
		name := CipherSuiteName(id)
		if !bytes.Contains(output, []byte(name)) {
			t.Errorf("cipher suite %s not found in PowerShell output", name)
		}
	}
}

func TestCipherSuitePreferenceSchannelTLS12(t *testing.T) {
	// Schannel should prefer AES-256 over AES-128.
	serverConfig := &Config{
		CipherSuites: []uint16{TLS_RSA_WITH_RC4_128_SHA, TLS_ECDHE_RSA_WITH_AES_128_GCM_SHA256},
		Certificates: testConfig.Certificates,
		MaxVersion:   VersionTLS12,
		GetConfigForClient: func(chi *ClientHelloInfo) (*Config, error) {
			if chi.CipherSuites[0] != TLS_ECDHE_RSA_WITH_AES_128_GCM_SHA256 {
				t.Error("the advertised order should not depend on Config.CipherSuites")
			}
			return nil, nil
		},
	}
	clientConfig := &Config{
		CipherSuites:       []uint16{TLS_RSA_WITH_RC4_128_SHA, TLS_ECDHE_RSA_WITH_AES_128_GCM_SHA256},
		InsecureSkipVerify: true,
	}
	state, _, err := testHandshake(t, clientConfig, serverConfig)
	if err != nil {
		t.Fatalf("handshake failed: %s", err)
	}
	if state.CipherSuite != TLS_ECDHE_RSA_WITH_AES_128_GCM_SHA256 {
		t.Error("the preference order should not depend on Config.CipherSuites")
	}
}

func TestCipherSuitePreferenceSchannelTLS13(t *testing.T) {
	// Schannel should prefer AES-256 over AES-128.
	serverConfig := &Config{
		Certificates: testConfig.Certificates,
		MaxVersion:   VersionTLS13,
	}
	clientConfig := &Config{
		InsecureSkipVerify: true,
	}
	state, _, err := testHandshake(t, clientConfig, serverConfig)
	if err != nil {
		t.Fatalf("handshake failed: %s", err)
	}
	if state.CipherSuite != TLS_AES_256_GCM_SHA384 {
		t.Error("the preference order should not depend on Config.CipherSuites")
	}
}
